// ============================================================================
// Project:	Deductive Database
// Filename:	list.h
// Purpose:	List of objects/tuples - relation with binding pattern f..f
// Last Change:	31.10.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2010-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	LIST_INCLUDED
#define LIST_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef STR_INCLUDED
#include "str.h"
#endif

#ifndef ERR_INCLUDED
#include "err.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef PAGE_INCLUDED
#include "page.h"
#endif

#ifndef MPOOL_INCLUDED
#include "mpool.h"
#endif

#ifndef REL_INCLUDED
#include "rel.h"
#endif


//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// LIST_MAGIC: Magic number (identifies objects of this class).
//-----------------------------------------------------------------------------

static const long LIST_MAGIC = 0x4C53540AL;	// 'LST\n'

//-----------------------------------------------------------------------------
// Null Pointers:
//-----------------------------------------------------------------------------

#define LIST_LEAF_NULL (static_cast<T*>(PAGE_NULL))
#define LIST_MID_NULL (static_cast<T**>(PAGE_NULL))
#define LIST_TOP_NULL (static_cast<T***>(PAGE_NULL))

//-----------------------------------------------------------------------------
// Forward declaration of the corresponding cursor class (for friend decl.):
//-----------------------------------------------------------------------------

template<class T> class cur_list_c;

//=============================================================================
// Class Template for Sequences/Lists (with Insertion only at the End):
//=============================================================================

template<class T> class list_c: public rel_c {
	public:

//-----------------------------------------------------------------------------
// Constructor, Destructor:
//-----------------------------------------------------------------------------

	// Constructor:
	list_c(str_t list_name) : rel_c(list_name, 0, T::NUM_COLS,
					"Tree of Tuple Arrays") {

		// Check parameter:
		CHECK(list_name != STR_NULL,
				"list_c::constructor: list_name is NULL");

		// Initialize attributes:
		Height      = 0;
		LeafNode    = LIST_LEAF_NULL;
		MidNode     = LIST_MID_NULL;
		TopNode     = LIST_TOP_NULL;
		LeafPtr     = LIST_LEAF_NULL;
		MidPtr      = LIST_MID_NULL;
		TopPtr      = LIST_TOP_NULL;
		LeafFree    = 0;
		MidFree     = 0;
		TopFree     = 0;
		UnusedPage1 = PAGE_NULL;
		UnusedPage2 = PAGE_NULL;

		// Set Magic number:
		CHECK_CODE(Magic = LIST_MAGIC);
	}

	// Destructor:
	~list_c()
	{
		// Check this object:
		CHECK_VALID("list_c::destructor");

		// Make this object invalid:
		CHECK_CODE(Magic = 0);
	}

//-----------------------------------------------------------------------------
// Copy-constructor and assignment operator are not supported for this class:
//-----------------------------------------------------------------------------

	private:

	list_c(const list_c& obj);		// Not implemented
	list_c& operator=(const list_c& obj);	// Not implemented

//=============================================================================
// Auxiliary Methods for Memory Page Management:
//=============================================================================

	private:

	// This handles the following problem:
	// When a memory page allocation fails,
	// we want to restore the state before the operation.
	// However, insert might have to allocate three memory pages,
	// and if the last allocation fails, there are two unused memory pages.
	// This is not a real memory leak, because the pages are stored
	// in the memory pool, and the memory pool will free them
	// when it is destroyed.
	//
	// It is also unlikely that a memory page allocation fails,
	// and later it works again (in which case every insert might
	// allocate two unused pages in the worst case).
	//
	// However, to be very correct, we remember the unused pages
	// in this object, and use them in the next memory allocation,
	// before we allocate new pages from the memory pool.
	// In this way, there can never be more than two unused pages
	// in this object.

//-----------------------------------------------------------------------------
// alloc_page: Get a memory page.
//-----------------------------------------------------------------------------

	inline page_t alloc_page() {
		if(UnusedPage1 == PAGE_NULL)
			return MemPool.alloc_page();
		page_t page = UnusedPage1;
		UnusedPage1 = UnusedPage2;
		UnusedPage2 = PAGE_NULL;
		return page;
	}

//-----------------------------------------------------------------------------
// free_page: De-allocate a memory page.
//-----------------------------------------------------------------------------

	inline void free_page(page_t page) {
		CHECK(page != PAGE_NULL, "list_c::free_page: page is NULL");
		if(UnusedPage1 == PAGE_NULL)
			UnusedPage1 = page;
		else {
			CHECK(UnusedPage2 == PAGE_NULL,
				"list_c::free_page: too many pages");
			UnusedPage2 = page;
		}
	}

//=============================================================================
// (Object) Methods:
//=============================================================================

	public:

//-----------------------------------------------------------------------------
// insert: Append new element to the list (at the end).
//-----------------------------------------------------------------------------

	inline bool insert(T elem) {
		page_t page;

		// Ensure that object is valid:
		CHECK_VALID("list_c::insert");

		// Most common case: Add elem to current leaf node:
		if(LeafFree > 0) {
			LeafFree--;
			*LeafPtr++ = elem;
			NumRows++;
			return true;
		}

		// Allocate first leaf:
		if(LeafNode == LIST_LEAF_NULL) {
			CHECK(Height == 0,
				"list_c::insert: null leaf, height != 0");
			CHECK(NumRows == 0,
				"list_c::insert: null leaf, num_rows != 0");
			page = alloc_page();
			if(page == PAGE_NULL) {
				MemErr = true;
				return false;
			}
			LeafNode = reinterpret_cast<T*>(page);
			LeafFree = PAGE_SIZE(T) - 1;
			CHECK(LeafFree >= 0, "list_c::insert: T too large");
			LeafPtr = LeafNode;
			*LeafPtr++ = elem;
			Height = 1;
			NumRows = 1;
			return true;
		}

		// There is a leaf node, but it is full:
		CHECK(LeafNode != LIST_LEAF_NULL,
			"list_c::insert: LeafNode is NULL");
		CHECK(LeafFree == 0,
			"list_c::insert: LeafFree not 0");

		// Remember old leaf node in case there is no mid node:
		T* old_leaf_node = LeafNode;

		// Allocate new leaf node:
		page = alloc_page();
		if(page == PAGE_NULL) {
			MemErr = true;
			return false;
		}
		LeafNode = reinterpret_cast<T*>(page);
		LeafFree = PAGE_SIZE(T) - 1;
		CHECK(LeafFree >= 0,
			"list_c::insert: T too large (2)");
		LeafPtr = LeafNode;
		*LeafPtr++ = elem;

		// Now we must enter the new leaf node to the mid branch level
		// In case there was no mid branch level before, both nodes,
		// the old leaf node and the new leaf node must be entered
		// to the first mid level branch node.

		// Construct first mid level branch node:
		if(MidNode == LIST_MID_NULL) {
			page = alloc_page();
			if(page == PAGE_NULL) {
				// Restore state before failed insert:
				free_page(LeafNode);
				LeafNode = old_leaf_node;
				LeafFree = 0;
				MemErr = true;
				return false;
			}
			MidNode = reinterpret_cast<T**>(page);
			MidFree = PAGE_SIZE(T*) - 2;
			CHECK(MidFree >= 0,
				"list_c::insert: T* too large");
			MidPtr = MidNode;
			*MidPtr++ = old_leaf_node;
			*MidPtr++ = LeafNode;
			Height = 2;
			NumRows++;
			return true;
		}

		// There is a mid level branch node and it is not yet full:
		else if(MidFree > 0) {
			MidFree--;
			*MidPtr++ = LeafNode;
			NumRows++;
			return true;
		}

		// There is a mid level branch node, but it is full:
		CHECK(MidNode != LIST_MID_NULL,
			"list_c::insert: MidNode is NULL");
		CHECK(MidFree == 0,
			"list_c::insert: MidFree not 0");

		// Construct additional (not first) mid level branch node:
		page = alloc_page();
		if(page == PAGE_NULL) {
			// Restore state before failed insert:
			free_page(LeafNode);
			LeafNode = old_leaf_node;
			LeafFree = 0;
			MemErr = true;
			return false;
		}
		T** old_mid_node = MidNode;
		MidNode = reinterpret_cast<T**>(page);
		MidFree = PAGE_SIZE(T*) - 1;
		CHECK(MidFree >= 0,
			"list_c::insert: T* too large (2)");
		MidPtr = MidNode;
		*MidPtr++ = LeafNode;

		// Now we must enter the new mid node to top level
		// (because it is not the first / only one).
		// Simple case: There is already a top node.
		if(TopNode != LIST_TOP_NULL) {
			// If it has no space, we would need another level,
			// which is not implemented:
			if(TopFree <= 0) {
				free_page(LeafNode);
				LeafNode = old_leaf_node;
				LeafFree = 0;
				free_page(MidNode);
				MidNode = old_mid_node;
				MidFree = 0;
				err_c::list_capacity_limit(Name, NumRows);
				return false;
			}
			TopFree--;
			*TopPtr++ = MidNode;
			NumRows++;
			return true;
		}

		// Finally handle the case that there is no top level node yet.
		// Construct one (there will always be only one):
		CHECK(TopNode == LIST_TOP_NULL,
				"list_c::insert: TopNode is not null");
		page = alloc_page();
		if(page == PAGE_NULL) {
			// Restore state before failed insert:
			free_page(LeafNode);
			LeafNode = old_leaf_node;
			LeafFree = 0;
			free_page(MidNode);
			MidNode = old_mid_node;
			MidFree = 0;
			MemErr = true;
			return false;
		}
		TopNode = reinterpret_cast<T***>(page);
		TopFree = PAGE_SIZE(T**) - 2;
		CHECK(TopFree >= 0,
				"list_c::insert: T** too large");
		TopPtr = TopNode;
		*TopPtr++ = old_mid_node;
		*TopPtr++ = MidNode;
		Height = 3;
		NumRows++;
		return true;
	}


//=============================================================================
// Debugging Support:
//=============================================================================

//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity check:
	public:
	str_t	check() const {

		// Check magic number:
		if(Magic != LIST_MAGIC)
			return "wrong magic number";

		// Check superclass:
		str_t msg = rel_c::check();
		if(msg)
			return msg;

		// Check number of free places in nodes:
		if(LeafFree < 0)
			CHECK_ERR("LeafFree is negative");
		if(MidFree < 0)
			CHECK_ERR("MidFree is negative");
		if(TopFree < 0)
			CHECK_ERR("TopFree is negative");
		if(LeafFree >= PAGE_SIZE(T))
			CHECK_ERR("LeafFree too large");
		if(MidFree >= PAGE_SIZE(T*))
			CHECK_ERR("MidFree too large");
		if(TopFree >= PAGE_SIZE(T**))
			CHECK_ERR("TopFree too large");

		// If a node is null, the number of free places must be 0:
		if(LeafNode == LIST_LEAF_NULL && LeafFree != 0)
			CHECK_ERR("LeafNode is null, but LeafFree != 0");
		if(MidNode == LIST_MID_NULL && MidFree != 0)
			CHECK_ERR("MidNode is null, but MidFree != 0");
		if(TopNode == LIST_TOP_NULL && TopFree != 0)
			CHECK_ERR("TopNode is null, but TopFree != 0");

		// Node and pointer can only be together null:
		if(LeafNode == LIST_LEAF_NULL && LeafPtr != LIST_LEAF_NULL)
			CHECK_ERR("LeafNode is null, but LeafPtr is not");
		if(LeafNode != LIST_LEAF_NULL && LeafPtr == LIST_LEAF_NULL)
			CHECK_ERR("LeafNode is not null, but LeafPtr is null");
		if(MidNode == LIST_MID_NULL && MidPtr != LIST_MID_NULL)
			CHECK_ERR("MidNode is null, but MidPtr is not");
		if(MidNode != LIST_MID_NULL && MidPtr == LIST_MID_NULL)
			CHECK_ERR("MidNode is not null, but MidPtr is null");
		if(TopNode == LIST_TOP_NULL && TopPtr != LIST_TOP_NULL)
			CHECK_ERR("TopNode is null, but TopPtr is not");
		if(TopNode != LIST_TOP_NULL && TopPtr == LIST_TOP_NULL)
			CHECK_ERR("TopNode is not null, but TopPtr is null");

		// Check that pointers correspond to free places:
		if(LeafNode != LIST_LEAF_NULL &&
			LeafFree != PAGE_SIZE(T) - (LeafPtr-LeafNode))
			CHECK_ERR("LeafFree and LeafPtr are inconsistent");
		if(MidNode != LIST_MID_NULL &&
			MidFree != PAGE_SIZE(T*) - (MidPtr-MidNode))
			CHECK_ERR("MidFree and MidPtr are inconsistent");
		if(TopNode != LIST_TOP_NULL &&
			TopFree != PAGE_SIZE(T**) - (TopPtr-TopNode))
			CHECK_ERR("TopFree and TopPtr are inconsistent");

		// Check by height:
		switch(Height) {
			case 0:
				if(NumRows != 0)
					CHECK_ERR("Height 0 - NumRows != 0");
				if(LeafNode != LIST_LEAF_NULL)
					CHECK_ERR("Height 0 - Leaf not null");
				if(MidNode != LIST_MID_NULL)
					CHECK_ERR("Height 0 - Mid not null");
				if(TopNode != LIST_TOP_NULL)
					CHECK_ERR("Height 0 - Top not null");
				if(LeafFree != 0)
					CHECK_ERR("Height 0 - LeafFree != 0");
				if(MidFree != 0)
					CHECK_ERR("Height 0 - MidFree != 0");
				if(TopFree != 0)
					CHECK_ERR("Height 0 - TopFree != 0");
				break;
			case 1:
				if(NumRows > PAGE_SIZE(T))
					CHECK_ERR(
					"Height 1 - NumRows too large");
				if(LeafNode == LIST_LEAF_NULL)
					CHECK_ERR("Height 1 - Leaf is null");
				if(MidNode != LIST_MID_NULL)
					CHECK_ERR("Height 1 - Mid not null");
				if(TopNode != LIST_TOP_NULL)
					CHECK_ERR("Height 1 - Top not null");
				if(MidFree != 0)
					CHECK_ERR("Height 1 - MidFree != 0");
				if(TopFree != 0)
					CHECK_ERR("Height 1 - TopFree != 0");
				break;
			case 2:
				if(NumRows > PAGE_SIZE(T) * PAGE_SIZE(T*))
					CHECK_ERR(
					"Height 2 - NumRows too large");
				if(LeafNode == LIST_LEAF_NULL)
					CHECK_ERR("Height 2 - Leaf is null");
				if(MidNode == LIST_MID_NULL)
					CHECK_ERR("Height 2 - Mid is null");
				if(TopNode != LIST_TOP_NULL)
					CHECK_ERR("Height 2 - Top not null");
				if(TopFree != 0)
					CHECK_ERR("Height 2 - TopFree != 0");
				break;
			case 3:
				if(NumRows > PAGE_SIZE(T) * PAGE_SIZE(T*)
						* PAGE_SIZE(T**))
					CHECK_ERR(
					"Height 3 - NumRows too large");
				if(LeafNode == LIST_LEAF_NULL)
					CHECK_ERR("Height 3 - Leaf is null");
				if(MidNode == LIST_MID_NULL)
					CHECK_ERR("Height 3 - Mid is null");
				if(TopNode == LIST_TOP_NULL)
					CHECK_ERR("Height 3 - Top is null");
				break;
			default:
				CHECK_ERR("Invalid height");
		}

		// Ok:
		return STR_NULL;
	}

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "LIST_MAGIC".
#endif

//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP

	public:

void dump(str_t headline = STR_NULL) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Relation/Sequence";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("Name:         '");
	err_c::dump_str(Name);
	err_c::dump_str("'");
	err_c::dump_nl();
	err_c::dump_str("NumRows:        ");
	err_c::dump_int(NumRows);
	err_c::dump_nl();
	err_c::dump_str("Height:        ");
	err_c::dump_int(Height);
	err_c::dump_nl();
	err_c::dump_str("NumPages:      ");
	err_c::dump_int(MemPool.num_alloc_pages());
	err_c::dump_nl();

	err_c::dump_str("Page Size:     ");
	err_c::dump_int(VER_PAGESIZE);
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_str("PAGE_SIZE(T):  ");
	err_c::dump_int(PAGE_SIZE(T));
	err_c::dump_nl();
	err_c::dump_str("PAGE_SIZE(T*): ");
	err_c::dump_int(PAGE_SIZE(T*));
	err_c::dump_nl();

	err_c::dump_str("LeafFree:      ");
	err_c::dump_int(LeafFree);
	err_c::dump_nl();
	err_c::dump_str("MidFree:       ");
	err_c::dump_int(MidFree);
	err_c::dump_nl();
	err_c::dump_str("TopFree:       ");
	err_c::dump_int(TopFree);
	err_c::dump_nl();

	err_c::dump_str("LeafNode:      ");
	err_c::dump_ptr(LeafNode);
	err_c::dump_nl();
	err_c::dump_str("LeafPtr:       ");
	err_c::dump_ptr(LeafPtr);
	err_c::dump_str(" [index ");
	err_c::dump_int(static_cast<int>(LeafPtr - LeafNode));
	err_c::dump_str("]");
	err_c::dump_nl();

	err_c::dump_str("MidNode:       ");
	err_c::dump_ptr(MidNode);
	err_c::dump_nl();
	err_c::dump_str("MidPtr:        ");
	err_c::dump_ptr(MidPtr);
	err_c::dump_str(" [index ");
	err_c::dump_int(static_cast<int>(MidPtr - MidNode));
	err_c::dump_str("]");
	err_c::dump_nl();

	err_c::dump_str("TopNode:       ");
	err_c::dump_ptr(TopNode);
	err_c::dump_nl();
	err_c::dump_str("TopPtr:        ");
	err_c::dump_ptr(TopPtr);
	err_c::dump_str(" [index ");
	err_c::dump_int(static_cast<int>(TopPtr - TopNode));
	err_c::dump_str("]");
	err_c::dump_nl();
	err_c::dump_nl();

	// Done:
	err_c::dump_end();
}

#endif


//=============================================================================
// Private Object Members:
//=============================================================================

	private:

	// Height: Height of the radix tree implementing the flexible array.
	int Height;	// Can only be 0 (empty tree), 1 (only leaf), 2, 3.

	// LeafNode: Current/Rightmost Leaf Node
	T *LeafNode;	// If Height == 1, this is the root.

	// MidNode: Current/Rightmost branch node one level above leaf nodes.
	T **MidNode;	// If Height == 2, this is the root.

	// TopNode: Branch node two levels above leaf nodes.
	T ***TopNode;	// Root for Height == 3. Otherwise unused.

	// LeafPtr: Next free entry in current leaf node.
	T *LeafPtr;

	// MidPtr: Next free entry in current branch node one above leaf nodes.
	T **MidPtr;

	// TopPtr: Next free entry in TopNode (two levels above leaf nodes).
	T ***TopPtr;

	// LeafFree: Free entries in current leaf node.
	int LeafFree;

	// MidFree: Free entries in current mid level node.
	int MidFree;

	// TopFree: Free entries in top level node.
	int TopFree;

	// Unused pages left over from a failed insert operation:
	page_t UnusedPage1;
	page_t UnusedPage2;

//-----------------------------------------------------------------------------
// This class works together with the corresponding cursor class:
//-----------------------------------------------------------------------------

	friend class cur_list_c<T>;

//=============================================================================
// End of Class Template:
//=============================================================================

};


//=============================================================================
// End of Include File:
//=============================================================================

#endif

